/* $Id: cache_failed_lookups.c,v 1.9 1997/07/11 23:59:39 gjb Exp $
 * cache_failed_lookups.c
 * (C) 1997 Greg J. Badros -- All Rights Reserved
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA
 * http://www.cs.washington.edu/homes/gjb
 */

#define NFS_NEED_XDR_TYPES
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>
/*#include <linux/errno.h> */

#include <asm/segment.h>
#include <asm/system.h>

#include "async_lookup.h"
#include "cache_failed_lookups.h"

#if 0
/* #define DEBUG_MOVING */
/* #define DEBUG_QUEUE */
#define DEBUG_RESET_TARGET
#define DEBUG_RECENT_HITS
/* #define DEBUG_RECENT_FAILS */
#define DEBUG_ADDING
#define DEBUG_PREDICTIONS
/* #define DEBUG_PRED_EXPIRE */
/* #define DEBUG_PRED_REPEAT */
#define DEBUG_SHOW_PREDICTIONS
/* #define DEBUG_INC_DEC */
/* #define DEBUG_FLT_FIND */
#endif

static unsigned long nfs_fl_timeout = .20 * HZ;

/* a queue window of the last failed lookups, uses cyclic-array impl.,
   lookups added to the tail, removed from the head */
static nfs_failed_lookup rgNFS_failures[NFS_FL_WINDOW_SIZE+1];
static int iflc_hd = 0;
static int iflc_tl = 0;

/* list of the directories and their transitions for the lookups in the
   above rgNFS_failures queue window.  These are in decreasing
   order of frequency (we do linear lookup, so we want dirs occurring
   most often to be cheapest to find */
static nfs_transitions rgNFS_transitions[NFS_FL_MAX_TOPLEVEL_DIRS+1];
static int cNFS_transitions = 0; /* # elem in above list */

static nfs_pred_lookup rgNFS_predictions[NFS_PRED_WINDOW_SIZE+1];
static int ip_hd = 0;
static int ip_tl = 0;


int cPredMadeTotal = 0;
int cPredUnused = 0;
int cPredCorrect = 0;
int cPredHits = 0;
int cRecentUseHit = 0;
int cTotLookups = 0;
int cPredOverflows = 0;

static int /* index into rgNFS_transitions[it].rg[] */
nfs_flt_target_find( int it, nfs_dirid dirid, int fHaveToFind )
{
  int i = 0;
  while (rgNFS_transitions[it].rg[i].dirid.dev != 0 &&
	 i < NFS_FL_MAX_TRANSITION_DIRS)
    {
    if (rgNFS_transitions[it].rg[i].dirid.dev == dirid.dev &&
	rgNFS_transitions[it].rg[i].dirid.ino == dirid.ino)
      return i;
    i++;
    }
  /* fHaveToFind is just there for debugging */
#ifndef NDEBUG_NFS_FL
  if (!fHaveToFind)
    printk("NFS: %s, could not find target!\n", __FUNCTION__);
#endif

  if (i == NFS_FL_MAX_TRANSITION_DIRS)
    { /* it's full, so cycle the spots with 1 counts, if any */
#ifndef NDEBUG_NFS_FL
    /* FIXGJB */
    printk("NFS: %s, Filled... ignoring\n",__FUNCTION__);
#endif
    return -1;
    }

  /* not full, so just add this to the end */
  rgNFS_transitions[it].rg[i].dirid = dirid;
  rgNFS_transitions[it].rg[i].c = 0;
  return i;
}

static int /* index into rgNFS_transitions */
nfs_flt_find(nfs_dirid dirid, int enumHaveToFind)
{ /* add an entry if it's not in there, and inc cNFS_transitions */
  int i = 0;
  for (;i<cNFS_transitions; i++)
    {
    /* FIXGJB: can't test structures for equality?  but can use assignment? */
    if (rgNFS_transitions[i].dirid.dev == dirid.dev &&
	rgNFS_transitions[i].dirid.ino == dirid.ino)
#ifdef DEBUG_FLT_FIND
      printk("NFS: %s found at %d\n",__FUNCTION__,i);
#endif
      return i;
    }
  if (enumHaveToFind == NO_CREATE)
    return -1;
  /* fHaveToFind is just there for debugging */
#ifndef NDEBUG_NFS_FL
  if (enumHaveToFind != NEED_NOT_FIND) 
    printk("NFS: %s, could not find entry!\n", __FUNCTION__);
#endif
  
  /* not found, so add it */
#ifdef DEBUG_ADDING
  printk("NFS_FLC: Adding (%d,%ld)\n",dirid.dev,dirid.ino);
#endif
  rgNFS_transitions[cNFS_transitions].dirid = dirid;
  rgNFS_transitions[cNFS_transitions].c = 0;
  memset(rgNFS_transitions[cNFS_transitions].rg, 0, 
	 NFS_FL_MAX_TOPLEVEL_DIRS * sizeof(nfs_transition_nextdir));
  cNFS_transitions++;
  return i;
}

static void
nfs_queue_async_lookup(struct nfs_dirid dirid, char *name)
{
  kdev_t dev = dirid.dev;
  unsigned long ino = dirid.ino;
  struct inode *dir = dirid.inode;
  int i;
  /* FIXGJB: this stuff should get done in lookup_result
     so it doesn't slow the client down any */
  /* expire old entries */
  while (ip_hd != ip_tl)
    {
    if (rgNFS_predictions[ip_hd].jiffies < jiffies - nfs_fl_timeout &&
	rgNFS_predictions[ip_hd].fValid)
      {
#ifdef DEBUG_PRED_EXPIRE
      printk("NFS_FLC: Expriring (%d,%ld) %s @ %ld, now %ld\n",
	     rgNFS_predictions[ip_hd].dirid.dev,
	     rgNFS_predictions[ip_hd].dirid.ino,
	     rgNFS_predictions[ip_hd].name,
	     rgNFS_predictions[ip_hd].jiffies,
	     jiffies);
#endif
      if (rgNFS_predictions[ip_hd].fUsed == FALSE)
	cPredUnused++;
      }
      ip_hd = QUEUE_INC(ip_hd,NFS_PRED_WINDOW_SIZE);
    }

  i = ip_hd;
  while (i != ip_tl)
    {
    if (rgNFS_predictions[i].dirid.dev == dev &&
	rgNFS_predictions[i].dirid.ino == ino &&
	(strncmp(rgNFS_predictions[i].name,name,NFS_FL_MAX_NAME_LEN) == 0))
      {
      /* already on the list, so we don't need to do anything, perhaps
	 if it's about to time out we could refresh it*/
#ifdef DEBUG_PRED_REPEAT
      printk("NFS_FLC: Repeat prediction\n");
#endif
      return;
      }
    i = QUEUE_INC(i,NFS_PRED_WINDOW_SIZE);
    }

  if (QUEUE_FULL(ip_hd,ip_tl,NFS_PRED_WINDOW_SIZE))
    {
    cPredOverflows++;
    return;
    }


#ifdef DEBUG_SHOW_PREDICTIONS
  printk("NFS_FLC: PREDICT: (%d,%ld) %s @ %ld\n",
	  dev,ino,name,jiffies);
#endif
  if (dev != dir->i_dev || ino != dir->i_ino)
    {
    /* the inode doesn't match the other info */
    printk("NFS_FLC: %s, non-matching inode (%d,%ld) vs. (%d,%ld)\n",__FUNCTION__,
	   dev,ino,dir->i_dev,dir->i_ino);
    return;
    }
  { /* scope */
  nfs_pred_lookup *ppred = &rgNFS_predictions[ip_tl];
  cPredMadeTotal++;
  ppred->dirid.dev = dev;
  ppred->dirid.ino = ino;
  ppred->dirid.inode = dir;
  ppred->jiffies = jiffies;
  ppred->fValid = FALSE;
  ppred->fUsed = FALSE;
  strncpy(ppred->name,name,NFS_FL_MAX_NAME_LEN);
  ip_tl = QUEUE_INC(ip_tl,NFS_PRED_WINDOW_SIZE);
  nfs_async_lookup(dir,name);
  } /* end scope */
}

/* was called nfs_predicted_right */
/* returns -index if found but not valid,
 * TRUE if found and valid
 * and FALSE if not found */
static int
nfs_in_flc(kdev_t dev, unsigned long ino, const char *name)
{
  /* this searches forwards, should it go backwards? probably
     does not matter, as the list is short, anyway */
  int i = ip_hd;
  while (i != ip_tl)
    {
    if (rgNFS_predictions[i].dirid.dev == dev &&
	rgNFS_predictions[i].dirid.ino == ino &&
	(strncmp(rgNFS_predictions[i].name,name,NFS_FL_MAX_NAME_LEN) == 0) &&
	rgNFS_predictions[i].jiffies + nfs_fl_timeout >= jiffies)
      {
      if (rgNFS_predictions[i].fValid == FALSE)
	{ /* we're still waiting on the lookup result */
	nfsflc_printk("still waiting");
	return -i;
	}
      if (rgNFS_predictions[i].fValid < 0)
	{
	nfsflc_printk("Found, but status was not -ENOENT");
	return FALSE;
	}
      if (rgNFS_predictions[i].fUsed == FALSE)
	{
	rgNFS_predictions[i].fUsed = TRUE;
	cPredCorrect++;
	}
      return TRUE;
      }
    i = QUEUE_INC(i,NFS_PRED_WINDOW_SIZE);
    }
  return FALSE;
}

int /* bool */
nfs_still_in_window(struct inode *dir, char *name)
{
  kdev_t dev = dir->i_dev;
  unsigned long ino = dir->i_ino;
  int i = QUEUE_DEC(iflc_tl,NFS_FL_WINDOW_SIZE);
  int c = 0;
  /* search backwards */
  do 
    {
    c++;
    if (rgNFS_failures[i].jiffies + nfs_fl_timeout < jiffies)
      break;
    if (rgNFS_failures[i].dirid.dev == dev &&
	rgNFS_failures[i].dirid.ino == ino &&
	(strncmp(rgNFS_failures[i].name,name,NFS_FL_MAX_NAME_LEN) == 0))
      {
      /* FIXGJB: do we need to check validity somehow, here? */
#ifdef DEBUG_RECENT_HITS
      printk("NFS_FLC: Reuse after %d checks\n",c);
#endif
      return TRUE;
      }
    i = QUEUE_DEC(i,NFS_FL_WINDOW_SIZE);
    }
  while (i != iflc_hd);
#ifdef DEBUG_RECENT_FAILS
  printk("NFS_FLC: %s failure after %d checks\n",__FUNCTION__,c);
#endif
  return FALSE;
}


int /* number of predictions made */
nfs_fl_make_predictions(struct inode *dir, char *name)
{ /* don't make predictions that we've already made */
  /* also pick some threshold only above-which we'll do the rpc call for the prediction */
  /* make no more than NFS_FL_MAX_PREDICTIONS per call */
  kdev_t dev = dir->i_dev;
  unsigned long ino = dir->i_ino;
  int it;
  int cPredMade = 0;
  int itarg = 0;
  int tot;
  nfs_dirid di;
  di.dev = dev; di.ino = ino;
  it = nfs_flt_find(di,NO_CREATE);
  if (it < 0)
    return 0;
  if ((tot = rgNFS_transitions[it].cTotTransitions) == 0)
    return 0;
  while (cPredMade < NFS_FL_MAX_PREDICTIONS)
    {
    int p = 100 * rgNFS_transitions[it].rg[itarg].c / tot;
    if (p > NFS_P_REQUIRED_TO_PREDICT) 
      {
      nfs_queue_async_lookup(rgNFS_transitions[it].rg[itarg].dirid,name);
      cPredMade++;
      } 
    else 
      break;
    itarg++;
    }
  return cPredMade;
}

static void
nfs_transitions_inc_and_order(int it,nfs_dirid diTarg)
{
  int iTarg;
  rgNFS_transitions[it].cTotTransitions++;
#ifdef DEBUG_INC_DEC
  printk("NFS_FLC: Inc spot %d, (%d,%ld), to (%d,%ld)\n",
	 it,rgNFS_transitions[it].dirid.dev,rgNFS_transitions[it].dirid.ino,
	 diTarg.dev,diTarg.ino);
#endif
  iTarg = nfs_flt_target_find(it, diTarg, NEED_NOT_FIND);
  rgNFS_transitions[it].rg[iTarg].c++;
  if (iTarg > 0 && 
      rgNFS_transitions[it].rg[iTarg].c > rgNFS_transitions[it].rg[iTarg-1].c)
    {
    nfs_transition_nextdir temp = rgNFS_transitions[it].rg[iTarg];
    int i = iTarg - 1;
    while (i > 0 &&
	   rgNFS_transitions[it].rg[iTarg].c > rgNFS_transitions[it].rg[i-1].c)
      i--;
    /* now swap */
    rgNFS_transitions[it].rg[iTarg] = rgNFS_transitions[it].rg[i];
    rgNFS_transitions[it].rg[i] = temp;
    }
}

static void
nfs_transitions_dec_and_order(int it,nfs_dirid diTarg)
{
  int iTarg;
  rgNFS_transitions[it].cTotTransitions--;
#ifdef DEBUG_INC_DEC
  printk("NFS_FLC: Dec spot %d, (%d,%ld), to (%d,%ld)\n",
	  it,rgNFS_transitions[it].dirid.dev,rgNFS_transitions[it].dirid.ino,
	  diTarg.dev,diTarg.ino);
#endif
  iTarg = nfs_flt_target_find(it, diTarg, MUST_FIND);
  if (--rgNFS_transitions[it].rg[iTarg].c == 0) 
    {
    rgNFS_transitions[it].rg[iTarg].dirid.dev = 0;
    rgNFS_transitions[it].rg[iTarg].dirid.ino = 0;
#ifdef DEBUG_RESET_TARGET
    printk("NFS_FLC: Reset target offset %d\n",iTarg);
#endif
    }

  while (iTarg < NFS_FL_MAX_TRANSITION_DIRS &&
	 rgNFS_transitions[it].rg[iTarg].c < rgNFS_transitions[it].rg[iTarg+1].c)
    {
    /* swap them */
    nfs_transition_nextdir temp = rgNFS_transitions[it].rg[iTarg];
#ifdef DEBUG_MOVING
    printk("NFS_FLC: Moving %d downto %d\n",iTarg+1,iTarg);
#endif
    rgNFS_transitions[it].rg[iTarg] = rgNFS_transitions[it].rg[iTarg+1];
    rgNFS_transitions[it].rg[iTarg+1] = temp;
    iTarg++;
#ifdef DEBUG_MOVING
    printk("NFS_FLC: Now %ld vs %ld\n",
	    rgNFS_transitions[it].rg[iTarg].c,
	    rgNFS_transitions[it].rg[iTarg+1].c);
#endif
    }
}

/* we pass in the index, since the fn needs to access the last several
   entries as well */
static void
nfs_fl_update(int iflc)
{
  int it = 0;
  /* update data structures for adding rgNFS_failures[iflc] */
#ifdef DEBUG_QUEUE
  printk("NFS_FLC: Updating %d\n",iflc);
#endif
  it = nfs_flt_find(rgNFS_failures[iflc].dirid,NEED_NOT_FIND);
  rgNFS_transitions[it].c++;
  /* see if we need to bubble it up */
  if (it > 0 && rgNFS_transitions[it].c > rgNFS_transitions[it-1].c)
    {
    nfs_transitions temp = rgNFS_transitions[it];
    int i = it - 1;
    while (i > 0 && rgNFS_transitions[it].c > rgNFS_transitions[i-1].c) i--;
    /* now swap them (not a stable ordering, but more efficient) */
    rgNFS_transitions[it] = rgNFS_transitions[i];
    rgNFS_transitions[i] = temp;
    it = i;
    }

  /* now we need to look through the rgNFS_failures before iflc (wrapping circularly)
     and increment the transition information to diAdd, this new directory */
  { /* scope */
  int iflcTarg = iflc;
  char *nameTarg = rgNFS_failures[iflcTarg].name;
  nfs_dirid diTarg = rgNFS_failures[iflcTarg].dirid;
  int i = 0;
  for ( ; i < NFS_FL_LOOKAHEAD; i++ )
    {
    iflc = QUEUE_DEC(iflc,NFS_FL_WINDOW_SIZE);
    /* if it's a lookup of the same name in a different directory, we need to increment */
    if ((strncmp(rgNFS_failures[iflc].name,nameTarg,NFS_FL_MAX_NAME_LEN) == 0) &&
	(rgNFS_failures[iflc].dirid.dev != diTarg.dev ||
	 rgNFS_failures[iflc].dirid.ino != diTarg.ino) &&
	rgNFS_failures[iflc].jiffies + nfs_fl_timeout >= rgNFS_failures[iflcTarg].jiffies)
      {
      int it = nfs_flt_find(rgNFS_failures[iflc].dirid,MUST_FIND);
      nfs_transitions_inc_and_order(it,diTarg);
      }
    }
  } /* end scope */
}

/* we pass in the index, since the fn needs to access the next several
   entries as well */
static void
nfs_fl_remove(int iflc)
{
  int it;
  /* update data structures for removing rgNFS_failures[iflc] */
#ifdef DEBUG_QUEUE
  printk("NFS_FLC: Removing %d\n",iflc);
#endif
  it = nfs_flt_find(rgNFS_failures[iflc].dirid,MUST_FIND);
  if (--rgNFS_transitions[it].c == 0)
    {
    cNFS_transitions--;
    /* and move the one from the last spot to here */
    rgNFS_transitions[it] = rgNFS_transitions[cNFS_transitions];
    rgNFS_transitions[cNFS_transitions].dirid.dev = 0;
    rgNFS_transitions[cNFS_transitions].dirid.ino = 0;
    rgNFS_transitions[cNFS_transitions].dirid.inode = NULL;
    memset(rgNFS_transitions[cNFS_transitions].rg, 0, 
	   NFS_FL_MAX_TOPLEVEL_DIRS * sizeof(nfs_transition_nextdir));
    return;
    }

  /* it has a non-zero count still */
  if (rgNFS_transitions[it].c < rgNFS_transitions[it+1].c)
    { /* bubble it down */
    nfs_transitions temp = rgNFS_transitions[it];
    int i = it+1;
    while (i+1 < cNFS_transitions &&
	   rgNFS_transitions[it].c < rgNFS_transitions[i+1].c) i++;
    /* now swap them (not a stable ordering, but more efficient) */
    rgNFS_transitions[it] = rgNFS_transitions[i];
    rgNFS_transitions[i] = temp;
    it = i;
    }

  /* now we need to look through the rgNFS_failures after iflc (wrapping circularly)
     and decrement the transition information from rgNFS_transitions[iflc].rg */
  { /* scope */
  int iflcSrc = iflc;
  char *name = rgNFS_failures[iflcSrc].name;
  int i = 0;
  for ( ; i < NFS_FL_LOOKAHEAD; i++)
    {
    iflc = QUEUE_INC(iflc,NFS_FL_WINDOW_SIZE);
    /* if it's a lookup of the same name in a different directory, we need to decrement */
    if (strncmp(rgNFS_failures[iflc].name,name,NFS_FL_MAX_NAME_LEN) == 0 &&
	(rgNFS_failures[iflc].dirid.dev != rgNFS_failures[iflcSrc].dirid.dev ||
	 rgNFS_failures[iflc].dirid.ino != rgNFS_failures[iflcSrc].dirid.ino) &&
	rgNFS_failures[iflc].jiffies <= rgNFS_failures[iflcSrc].jiffies + nfs_fl_timeout)
      {
      nfs_transitions_dec_and_order(it,rgNFS_failures[iflc].dirid);
      }
    }
  } /* end scope */
}

static void
nfs_fl_add(struct inode *dir, char *name)
{
  kdev_t dev = dir->i_dev;
  unsigned long ino = dir->i_ino;
  /* I could choose to not add things already in the cache,
     but that would take a linear scan of the list and doesn't seem
     to be worth the trouble */
  if (strlen(name) > NFS_FL_MAX_NAME_LEN)
    return;
  if (iflc_hd == QUEUE_INC(iflc_tl,NFS_FL_WINDOW_SIZE))
    { /* already full, so remove first */ 
    nfs_fl_remove(iflc_hd);
    iflc_hd = QUEUE_INC(iflc_hd,NFS_FL_WINDOW_SIZE);
    }
  rgNFS_failures[iflc_tl].dirid.dev = dev;
  rgNFS_failures[iflc_tl].dirid.ino = ino;
  rgNFS_failures[iflc_tl].dirid.inode = dir;
  strncpy(rgNFS_failures[iflc_tl].name,name,NFS_FL_MAX_NAME_LEN);
  rgNFS_failures[iflc_tl].jiffies = jiffies;
  /* now update data structure for adding rgNFS_failures[iflc_tl] */
  nfs_fl_update(iflc_tl);
  iflc_tl = QUEUE_INC(iflc_tl,NFS_FL_WINDOW_SIZE);
}


void
nfs_handle_failed_lookup(struct inode *dir,char *name,int error)
{
  cTotLookups++;
  nfs_fl_make_predictions(dir,name);
  nfs_fl_add(dir,name);
  return;
}

int
nfs_failed_lookup_cache_lookup(struct inode *dir, char *name)
{
  /* return the error from the cached lookup (-ENOENT),
     or 0 if not in cache */
  int status = nfs_in_flc(dir->i_dev,dir->i_ino,name);
  if (status == TRUE) /* must == TRUE */
    {
    cPredHits++;
#ifdef DEBUG_FAILED_LOOKUP_CACHE_HITS
    printk("NFS_FLC: Prediction HIT!\n");
#endif
    return -ENOENT;
    }
  else if (status == FALSE)
    {
    if (nfs_still_in_window(dir,name))
      {
      cRecentUseHit++;
#ifdef DEBUG_FAILED_LOOKUP_CACHE_HITS
      printk("NFS_FLC: Recent Use HIT!\n");
#endif
      return -ENOENT;
      }
    return 0;
    }
  else
    {
    int index = -status;
    int cLoops = 2;
    while (cLoops-- > 0)
      {
      current->timeout = jiffies + 1;
      current->signal = 0;
      current->state = TASK_INTERRUPTIBLE;
      schedule();
      if (rgNFS_predictions[index].fValid)
	return -ENOENT;
      }
    }
  printk("NFS: %s, waited, but still did not get valid\n",__FUNCTION__);
  return 0;
}

int
nfs_failed_lookup_result(kdev_t dev, unsigned long ino, const char *name, int status)
{
  int ipred = nfs_in_flc(dev,ino,name);
  if (ipred == TRUE)
    {
    nfsflc_printk("already valid");
    if (status != -ENOENT)
      printk("NFS: %s, already valid but status = %d, not ENOENT\n",__FUNCTION__,status);
    return 0;
    }
  if (ipred == FALSE)
    {
    nfsflc_printk("could not find prediction slot");
    if (status != -ENOENT)
      printk("NFS: %s, no slot but status = %d, not ENOENT\n",__FUNCTION__,status);
    return 0;
    }
  ipred = -ipred;
  /* set it valid */
  if (status != -ENOENT)
    {
    printk("NFS: %s, status = %d, not ENOENT\n",__FUNCTION__,status);
    rgNFS_predictions[ipred].fValid = status;
    }
  nfsflc_printk("setting valid");
  rgNFS_predictions[ipred].fValid = TRUE;
  return 0;
}
