#include <stdio.h>
#include <stdlib.h>

#define printk(x...) printf(x)
#define nfs_printk(x...) printf(x)
#define verb_nfs_printk(x...)
#define PAGE_SIZE 4096

struct inode {
	unsigned long i_size;
};

struct nfs_cache_info {

	struct inode *inodeCache;

	unsigned char *pchValidBitmap;

	/* Use cpages{Valid,Total} to tell when we have filled the bitmap */
	int cpagesValid;
	int cpagesTotal;

};

static inline
char *
nfs_kmalloc_bitmap(int num_pages)
{
  int num_bytes = (num_pages - 1)/8 + 1;
  char *pch = (char *) malloc(num_bytes);
	int i = 0;
  if (pch == NULL) {
		printk("NFS: nfs_kmalloc_bitmap -- could not allocate %d bytes\n",num_bytes);
  }
	nfs_printk("NFS: nfs_kmalloc_bitmap allocated %d bytes for %d pages\n",
						 num_bytes, num_pages);
	for (; i<num_bytes; i++)
		pch[i] = 0;
  return pch;
}

/* must kfree pnci->pchValidBitmap after this call */
nfs_initialize_nci(struct nfs_cache_info *pnci,
									 struct inode * inode, 
									 struct inode * inodeCache)
{
  int num_pages = ((long) inode->i_size - 1)/PAGE_SIZE + 1;
	pnci->inodeCache = inodeCache;
	pnci->pchValidBitmap = nfs_kmalloc_bitmap(num_pages);
	pnci->cpagesValid = 0;
	pnci->cpagesTotal = num_pages;
}

static inline
int /* non-zero is true */
nfs_is_bitmap_valid_for_ipage(const struct nfs_cache_info *const pnci,unsigned long ipage)
{
  int ibyte = ipage/8;
  int ibit = ipage - ibyte*8;
  int valid = pnci->pchValidBitmap[ibyte] & 1<<(7-ibit);
  verb_nfs_printk("NFS:nfs_is_bitmap_valid_for_ipage %lu (byte,bit) = (%d,%d) is %s\n",
						 ipage, ibyte, ibit,valid?"true":"false");
  return valid;
}

static inline
void
nfs_mark_bitmap_valid_for_ipage(struct nfs_cache_info *pnci, unsigned long ipage)
{
  int ibyte = ipage/8;
  int ibit = ipage - ibyte*8;
  nfs_printk("NFS:nfs_mark_bitmap_valid_for_ipage %lu (byte,bit) = (%d,%d)\n",
						 ipage, ibyte, ibit);
#ifndef NO_DEBUG_CACHE_BITMAP
	if (nfs_is_bitmap_valid_for_ipage(pnci,ipage)) {
		printk("NFS: nfs_mark_bitmap_valid_for_ipage already valid (byte,bit) = (%d,%d)\n",
					 ibyte, ibit);
		pnci->cpagesValid--;
	}
#endif
  pnci->pchValidBitmap[ibyte] |= 1<<(7-ibit);
	pnci->cpagesValid++;
}

static inline
int /* bool */
nfs_is_nci_complete(const struct nfs_cache_info *const pnci)
{
	return pnci->cpagesValid == pnci->cpagesTotal;
}

int
main(int argc, char *argv[])
{
	char szCommand[128];
	struct nfs_cache_info nci;
	struct inode a_inode;
  char chCommand = '\0';
  unsigned long number = 0;
	int f = 0;
	int i = 0;

	while (1) {
		gets(szCommand);
		sscanf(szCommand,"%c %d",&chCommand,&number);
		switch (toupper(chCommand)) {
			/* Make a new nfs_cache_info */
		case 'M':
			a_inode.i_size = number;
			nfs_initialize_nci(&nci,&a_inode,NULL);
			break;
			/* Check if a page number is valid */
		case 'C':
			f = nfs_is_bitmap_valid_for_ipage(&nci,number);
			break;
			/* Validate (set it valid) a page number */
		case 'V':
			nfs_mark_bitmap_valid_for_ipage(&nci,number);
			break;
			/* Dump the bytes */
		case 'D':
			printf("%d: ",nci.cpagesTotal);
			for (i = 0; i<(nci.cpagesTotal-1)/8+1; i++)
				printf("%x ",nci.pchValidBitmap[i]);
			printf("\n");
			break;
			/* List the bitmap as --*--*--*/
		case 'L':
			printf("%d: ",nci.cpagesTotal);
			for (i = 0; i<nci.cpagesTotal; i++)
				printf("%c",nfs_is_bitmap_valid_for_ipage(&nci,i)?'*':'-');
			printf("  %s\n",nfs_is_nci_complete(&nci)?"complete":"");
			break;
			/* eXit */
		case 'X':
			return 0;
		}
	}
}

